/********************************************************* begin of preamble
**
** Copyright (C) 2003-2012 Software- und Organisations-Service GmbH. 
** All rights reserved.
**
** This file may be used under the terms of either the 
**
**   GNU General Public License version 2.0 (GPL)
**
**   as published by the Free Software Foundation
**   http://www.gnu.org/licenses/gpl-2.0.txt and appearing in the file
**   LICENSE.GPL included in the packaging of this file. 
**
** or the
**  
**   Agreement for Purchase and Licensing
**
**   as offered by Software- und Organisations-Service GmbH
**   in the respective terms of supply that ship with this file.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
** IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
** THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
** BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
********************************************************** end of preamble*/
package com.sos.jobnet.creator;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.creator.jobs.JobNetCreatorOptions;
import com.sos.jobnet.db.JobNetPlanDBItem;
import com.sos.localization.Messages;
import com.sos.scheduler.model.SchedulerObjectFactory;
import com.sos.scheduler.model.objects.*;
import com.sos.scheduler.model.answers.Answer;
import com.sos.scheduler.model.answers.ERROR;
import com.sos.scheduler.model.commands.JSCmdAddOrder;

/**
 * \file JobNetPlanExecutor.java
 * \brief creates a bootstrap order for every jobnet created with JobNetPlanCreator 
 *  
 * \class JobNetPlanExecutor
 * \brief creates a bootstrap order for every jobnet created with JobNetPlanCreator 
 * 
 * \details
 * The JobNetPlan creation for the given bootstrap order, which is called first in the run method, results 
 * one or more jobnet plans. For each plan an order is created in the JobScheduler by sending the 
 * command ADD_ORDER. 
 * \code
  \endcode
 *
 * \author Stefan Schdlich
 * \version 1.0 - 03.04.2012 10:53:07
 * <div class="sos_branding">
 *   <p> 2010 SOS GmbH - Berlin (<a style='color:silver' href='http://www.sos-berlin.com'>http://www.sos-berlin.com</a>)</p>
 * </div>
 */
public class JobNetPlanExecutor {
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id$";
	
	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
	private final static String JOBNETC_E_0010 = "JOBNETC_E_0010"; // order not started: %1$s.
	private final static String JOBNETC_I_0011 = "JOBNETC_I_0011"; // order %1$s scheduled for %2$s.
	private final static String JOBNETC_E_0016 = "JOBNETC_E_0016"; // error creating the order for job net plan

	private final static Logger logger = Logger.getLogger(JobNetPlanExecutor.class);
	private final static String YES = "yes";
	
	private final ObjectModelConnector connection;
	private final SchedulerObjectFactory factory;
	private List<String> commands = new ArrayList<String>();
	
	private int count = 0;

	public JobNetPlanExecutor(JobNetCreatorOptions options) {
		connection = ObjectModelConnector.getInstance(options.SCHEDULER_HOST, options.SCHEDULER_TCP_PORT, options.SCHEDULER_CONFIGURATION_DIRECTORY);
		factory = connection.getFactory();
	}

	public void startOrder(JobNetPlanDBItem bootDBItem, JobNetOrder order) {
		
		try {
			// JobNetOrder order = getBootstrapOrder();
			JSCmdAddOrder cmdOrder = factory.createAddOrder();
			
			String id = getUniqueOrderId(order.getJobNetOrderName());
			String title = (order.getTitle() == null) ? "<no title>" : order.getTitle();
			
			logger.debug("create order for UUID " + bootDBItem.getUuid() + " with order id " + id );
			cmdOrder.setJobChain(order.getJobNetJobChainName());
			cmdOrder.setReplace(YES);
			cmdOrder.setId( id );
			cmdOrder.setTitle("JobNet: " + title);
			Param p1 = factory.createParam(JobNetConstants.UUID_PARAMETER_NAME,bootDBItem.getUuid());
			Param p2 = factory.createParam(JobNetConstants.JOB_NET_PARAMETER_NAME,order.getJobNetJobChainName());
			Param p3 = factory.createParam(JobNetConstants.ORDER_NAME_PARAMETER_NAME,order.getJobNetOrderName());
			cmdOrder.setParams(order.getParams());
			cmdOrder.getParams().getParamOrCopyParamsOrInclude().add(p1);
			cmdOrder.getParams().getParamOrCopyParamsOrInclude().add(p2);
			cmdOrder.getParams().getParamOrCopyParamsOrInclude().add(p3);
			String startAt = getStartTime(bootDBItem);
			cmdOrder.setAt(startAt);
			logger.debug("create order for UUID " + bootDBItem.getUuid() + " with order id " + id + " start at " + startAt );
			
			String xml = cmdOrder.toXMLString();
			commands.add(xml);
			cmdOrder.run();
			Answer answer = cmdOrder.getAnswer();
			ERROR jsError = answer.getERROR();
			if (jsError != null) {
				String msgText = msg.getMsg(JOBNETC_E_0010,jsError.getText());
				logger.error(msgText);
				logger.debug(xml);
				throw new JobNetException(msgText);
			}
			
			logger.info(msg.getMsg(JOBNETC_I_0011,id,startAt));
		} catch (Exception e) {
			String msgText = msg.getMsg(JOBNETC_E_0016);
			logger.error(msgText);
			throw new JobSchedulerException(msgText,e);
		}

	}
	
	private String getUniqueOrderId(String orderId) throws InterruptedException {
		String result = orderId;
		if (count != 0) { 
			Thread.sleep(100);	// to ensure that the order id is unique
			DateTimeFormatter fmtDate = DateTimeFormat.forPattern("eHHmmssSSS");
			result += "_" + fmtDate.print( new DateTime() );
		}
		count++;
		return result;
	}
	
	private String getStartTime(JobNetPlanDBItem bootStrapDBItem) {
    	SimpleDateFormat df = new SimpleDateFormat(bootStrapDBItem.dateFormat);
		return df.format(bootStrapDBItem.getPlannedStartTime());
	}

	public List<String> getCommands() {
		return commands;
	}

}
